import functools
from typing import List, Tuple

from potator.models import EntityData, File
from potator.tokenizer.buckwheat import get_identifiers_sequence_from_code, recognize_languages_dir
from potator.tokenizer.buckwheat.tokenizer import get_data_from_file
from potator.tokenizer.buckwheat.utils import FileData, ObjectData, transform_files_list, get_full_path


MIN_ENTITY_LENGTH = 128


class EntitiesExtractor:
    @staticmethod
    def extract_entity_from_object_data(obj: ObjectData, file_path: str) -> EntityData:
        entity_tokens = get_identifiers_sequence_from_code(obj.content, obj.lang)
        return EntityData(obj, entity_tokens, file_path)

    @staticmethod
    def extract_entity_from_file_data(file: FileData) -> List[EntityData]:
        return [EntitiesExtractor.extract_entity_from_object_data(obj, file.path)
                for obj in file.objects if obj.end_byte - obj.start_byte >= MIN_ENTITY_LENGTH]

    @staticmethod
    def extract_entities_from_files_data(files: List[FileData]) -> List[EntityData]:
        def acc(value: List[EntityData], file_data: FileData) -> List[EntityData]:
            value.extend(EntitiesExtractor.extract_entity_from_file_data(file_data))
            return value

        return functools.reduce(acc, files, [])

    @staticmethod
    def extract_data_from_files(files: List[File]) -> List[FileData]:
        return [get_data_from_file(file, lang, True, False) for file, lang in files]

    @staticmethod
    def extract_data_from_directory(directory: str, granularity: str)\
            -> Tuple[List[File], List[FileData], List[EntityData]]:
        lang2files = recognize_languages_dir(directory)
        files = transform_files_list(lang2files, granularity, None)
        files = [(get_full_path(file, directory), lang) for file, lang in files]
        files_data = EntitiesExtractor.extract_data_from_files(files)
        entities = EntitiesExtractor.extract_entities_from_files_data(files_data)
        return files, files_data, entities



from collections import Counter
from typing import List


def overlap(bag_of_tokens1: List[str], bag_of_tokens2: List[str]) -> float:
    """
    Given two bag of tokens return their overlap
    :param bag_of_tokens1:
    :param bag_of_tokens2:
    :return: value of overlap
    """
    return sum((Counter(bag_of_tokens1) & Counter(bag_of_tokens2)).values())


def jaccard(bag_of_tokens1: List[str], bag_of_tokens2: List[str]) -> float:
    """
    Given two bag of tokens return their jaccard similarity
    :param bag_of_tokens1:
    :param bag_of_tokens2:
    :return: value of overlap
    """
    max_len = max(len(bag_of_tokens1), len(bag_of_tokens2))
    if max_len == 0:
        return 0
    return overlap(bag_of_tokens1, bag_of_tokens2) / max_len



import os
from pathlib import Path


TOKENIZER_DIR = 'potator/tokenizer'
TOKENIZER_URL = 'https://github.com/otzhora/buckwheat'


def setup_tokenizer() -> None:
    """
    Clone GitHub repository with tokenizer and setup it.
    :return: None.
    """
    cwd = Path(os.getcwd())

    if not os.path.exists(cwd/TOKENIZER_DIR):
        os.system(f'git clone {TOKENIZER_URL} {TOKENIZER_DIR}')

    os.chdir(cwd/TOKENIZER_DIR)
    os.system(f'git pull')
    os.system('git submodule update --init --recursive --depth 1')
    os.chdir('..')

    from .tokenizer.buckwheat.parsing import main as parsing_main
    os.chdir(cwd/TOKENIZER_DIR/"buckwheat"/"parsing")
    parsing_main()

    from .tokenizer.buckwheat.language_recognition import main as enry_main
    os.chdir(cwd/TOKENIZER_DIR/"buckwheat"/"language_recognition")
    enry_main()
    os.chdir(cwd)


if __name__ == '__main__':
    setup_tokenizer()



import os
from collections import defaultdict, Counter
from difflib import HtmlDiff
from typing import List, Dict

from potator.models import DetectionResult, EntityData


HTML_HEADER = """
<head>
<meta http-equiv="Content-Type"
      content="text/html; charset=ISO-8859-1" />
<title></title>
<style type="text/css">
    table.diff {font-family:Courier; border:medium;}
    .diff_header {background-color:#e0e0e0}
    td.diff_header {text-align:right}
    .diff_next {background-color:#c0c0c0}
    .diff_add {background-color:#aaffaa}
    .diff_chg {background-color:#ffff77}
    .diff_sub {background-color:#ffaaaa}
</style>
</head>
"""


def get_formatted_detection_result(result: DetectionResult) -> str:
    differ = HtmlDiff()

    algorithm = f"Algorithm '{result.used_algorithm}' was used to get those results <br>"
    clones = result.clones
    tables = []
    for clone in clones:
        code1 = clone.entity_1.object_data.content.splitlines()
        code2 = clone.entity_2.object_data.content.splitlines()
        src1 = clone.entity_1.file_path
        src2 = clone.entity_2.file_path
        score = clone.similarity
        table = differ.make_table(code1, code2, fromdesc=f"{src1}, similarity: {score}", todesc=src2, context=True)
        tables.append(table)
        tables.append("<br>")
    return f"<html><body> {HTML_HEADER} {algorithm} {''.join(tables)} </body></html>"


def write_formatted_detection_result(result: DetectionResult, out: str) -> None:
    html = get_formatted_detection_result(result)
    with open(out, "w") as f:
        f.write(html)


def make_absolute_path(path: str) -> str:
    if os.path.isabs(path):
        return path
    return os.path.join(os.getcwd(), path)


LANGUAGE_ORDER = Dict[str, int]


def build_global_token_counts_from_entities(entities: List[EntityData]) -> Dict[str, LANGUAGE_ORDER]:
    global_token_counts = defaultdict(Counter)
    for entity in entities:
        global_token_counts[entity.object_data.lang] += Counter(entity.bag_of_tokens)

    return global_token_counts


def sort_tokens_gtc(entities: List[EntityData]) -> Dict[str, LANGUAGE_ORDER]:
    global_token_counts = build_global_token_counts_from_entities(entities)

    for entity in entities:
        lang_token_counts = global_token_counts[entity.object_data.lang]
        entity.bag_of_tokens.sort(key=lambda token: lang_token_counts[token])
    return global_token_counts



from setuptools import setup
from os import path
import pathlib

from potator.setup_tokenizer import setup_tokenizer


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

with open(HERE / 'requirements.txt', 'r', encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
    install_requires = [x.strip() for x in all_reqs if ('git+' not in x) and (not x.startswith('#'))
                        and (not x.startswith('-'))]
    dependency_links = [x.strip().replace('git+', '') for x in all_reqs if 'git+' not in x]

setup_tokenizer()

setup(
    name='potator',
    description='',
    version='0.1.1',
    packages=['potator'],
    package_dir={'potator': 'potator'},
    install_requires=install_requires,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'potator = potator.__main__:main'
        ]
    },
    author="Yuriy Rogachev",
    long_description=README,
    long_description_content_type="text/markdown",
    license='MIT',
    url='https://github.com/otzhora/potator',
    download_url='https://github.com/otzhora/potator/releases/tag/v0.1.1',
    dependency_links=dependency_links,
    keywords=['STATIC-ANALYSIS', 'PLAGIARISM-DETECTION', 'PLAGIARISM-DETECTOR'],
    author_email='rogachev.yuiry28@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'])



from dataclasses import dataclass
from typing import List, Tuple

from potator.tokenizer.buckwheat.utils import ObjectData


File = Tuple[str, str]


@dataclass
class EntityData:
    object_data: ObjectData
    bag_of_tokens: List[str]
    file_path: str

    def __hash__(self):
        return hash(f"{self.file_path} + {self.object_data.content}")


@dataclass
class CloneData:
    entity_1: EntityData
    entity_2: EntityData
    similarity: float

    def __lt__(self, other):
        return self.similarity < other.similarity


@dataclass
class DetectionResult:
    """
    Class for storing results of detection
    """
    clones: List[CloneData]
    used_algorithm: str
